/*
  KLIENT sieci wymiany plikow wg. protokolu grupy I4, z srody 8:00
  Jacek Pospychaa (66301)
 */

#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>
#include <sys/sem.h>
#include <sys/wait.h>

#include <regex.h>
#include <stdio.h>
#include <string.h>

#define H66301_CLIENT

#include "66301_EXTENDEDprotocol.h"
#include "66301_common.h"

#define PREFIX ""

#define CMD_NOOP  0
#define CMD_QUIT -1
#define CMD_HELP -2

#define BUFSIZE 1024

typedef struct client_mem {
  int shmid;
  int msgid;
  int serverid;
  int logged_in;
  pid_t childproc_pid;
  char* username;
  int extended;
} client_mem;

void logout(void);
void safe_exit(int);

/**************************************
 * Zarzdzanie pamici wspdzielon
 **************************************/

/* Wskanik do pamici wspdzielonej klienta

   Jeli argument!=NULL, zapamituje wskanik do wspdzielonej pamici.
   Jeli argument==NULL, zwraca wskanik do tej pamici.
   Jeli argument==-1, ustawia wskanik na NULL
   Dziaa podobnie do zmiennej globalnej, wykorzystywane przy safe_exit,
   by atwo znale obszar pamici wspdzielonej do zwolnienia.
 */
client_mem* shared_space_location(client_mem* pointer)
{
  static client_mem* mem_ptr = NULL;
  if (pointer != NULL)
    mem_ptr = pointer;
  if ((int)pointer == -1)
    mem_ptr = NULL;
  return mem_ptr;
}

/* Tworzy pami wspdzielon, to znaczy obszar do przechowywania zmiennych
   opisujcych stan uytkownika.
*/
client_mem* make_shared_space(void)
{
  DEBUGMSG(GENERAL_DBG, ("making shared space...\n"));
  int shmid = shmget(IPC_PRIVATE, sizeof(client_mem), 0644|IPC_CREAT);
  client_mem* mem = NULL;

  if (shmid == -1)
    {
      perror("make_shared_space shmget");
      safe_exit(SIGTERM);
    }

  mem = shmat(shmid, 0, 0);
  DEBUGMSG(MEM_DBG, ("attaching shared memory at %x\n", (unsigned int)mem));
  
  if ((int)mem == -1)
    {
      perror("make_shared_space shmat");
      safe_exit(SIGTERM);
    }
  bzero(mem, sizeof(client_mem));
  mem->shmid = shmid;
  mem->serverid = -1;

  return shared_space_location(mem);
}

/* Usuwa obszar pamici wspdzielonej. Ewentualne bdy s wywietlane,    
   ale nie powoduj zatrzymania programu.
 */
void remove_shared_space(void)
{
  DEBUGMSG(GENERAL_DBG, ("removing shared space...\n"));
  int shmid;
  client_mem* mem = shared_space_location(NULL);
  shared_space_location((client_mem*)-1);

  if (mem != NULL)
    {
      shmid = mem->shmid;

      if (shmdt(mem) == -1)
	{
	  perror("remove_shared_space shmdt");
	}
      if (shmctl(shmid, IPC_RMID, NULL) == -1)
	{
	  perror("remove_shared_space shmctl");
	}
    }
}

int ssend_message(int command, char* arg1, char* arg2, char* body)
{
  client_mem *mem = shared_space_location(NULL);
  complete_msg msg;
  ASSERT(mem!=NULL,);

  if (command == CMD_LOGIN)
    {
      msg.clientid = 1;
    }
  else
    {
      if (! mem->logged_in)
	{
	  printf("Najpierw musisz sie zalogowac\n");
	  return 1;
	}
      else
	{
	  msg.clientid = mem->msgid;
	}
    }

  msg.command = command;
  msg.argument1 = arg1;
  msg.argument2 = arg2;
  msg.messagebody = body;

  if (mem->serverid == -1)
    {
      printf("Brak poczenia z serwerem\n");
      return 1;
    }
  else
    {
      return send_message(mem->serverid, &msg);
    }
}

/**************************************
 * Bezpieczne wyjcie
 **************************************/

void safe_exit(int signal)
{
  DEBUGMSG(GENERAL_DBG, ("client process %d leaving\n", getpid()));
  client_mem *mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  if (getpid() != mem->childproc_pid)
    {
      if (mem->logged_in)
	{
	  logout();
	}
      kill(mem->childproc_pid, SIGTERM);
      waitpid(mem->childproc_pid, NULL, 0);
      close_queue(mem->msgid);
      mem->msgid = 0;
      remove_shared_space();
    }
  exit(0);
}

/**************************************
 * Polecenia
 **************************************/

void login(char *username)
{
  char clientid_str[BUFSIZE] = {0};
  client_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);

  if (mem->logged_in)
    {
      printf("Juz jestes zalogowany.\n");
    }
  else
    {
      if ((mem->serverid = msgget(SERVERID, 0)) == -1)
	{
	  printf("Nie mozna znalezc kolejki serwera. Czy serwer dziala?\n");
	}
      else
	{
	  snprintf(clientid_str, BUFSIZE, "%d", mem->msgid);
	  ssend_message(CMD_LOGIN, username, clientid_str, NULL);
	}
    }
}

void logout(void)
{
  //  client_mem* mem = shared_space_location(NULL);
  //  ASSERT(mem!=NULL,);
  ssend_message(CMD_LOGOUT, NULL, NULL, NULL);
}

void share(char *sharename, char *username, char *filename)
{
  FILE *f;
  char *filecontent;
  long fsize;
  client_mem* mem = shared_space_location(NULL);

  ASSERT(mem!=NULL,);

  if ((f = fopen(filename, "r")) == NULL)
    {
      printf("Nie mozna otworzyc pliku \"%s\"\n", filename);
    }
  else
    {
      fseek(f, 0, SEEK_END);
      fsize = ftell(f);
      fseek(f, 0, SEEK_SET);
      filecontent = xmalloc(fsize + 1);
      bzero(filecontent, fsize+1);
      fread(filecontent, 1, fsize, f);

      ssend_message(CMD_SHARE, sharename, username, filecontent);

      xfree(filecontent);
      fclose(f);
    }
}

void allow(char *sharename, char *username)
{
  ssend_message(ECMD_ALLOW, sharename, username, NULL);
}

void shutdown(void)
{
  ssend_message(ECMD_SHUTDOWN, NULL, NULL, NULL);
}

void kickall(void)
{
  ssend_message(ECMD_KICKALL, NULL, NULL, NULL);
}

void unshare(char *sharename)
{
  ssend_message(ECMD_UNSHARE, sharename, NULL, NULL);
}

void get(char *sharename, char *username)
{
  ssend_message(CMD_GET, username, sharename, NULL);
}

void list(char *username)
{
  ssend_message(CMD_LIST, username, NULL, NULL);
}

void who(void)
{
  ssend_message(CMD_WHO, NULL, NULL, NULL);
}

void stat(void)
{
  ssend_message(CMD_STAT, NULL, NULL, NULL);
}

void kick(char *username)
{
  ssend_message(CMD_KICK, username, NULL, NULL);
}

void set(char* varname, char* newvalue)
{
  ssend_message(CMD_SET, varname, newvalue, NULL);
}

void splash_screen(void)
{
  printf("66301_client: klient sieci IPC1500. Autor: Jacek Pospychala (inf66301)\n"
	 "Jesli nie wiesz jak zaczac, wpisz \"help\".\n\n"
	 "Porada dnia: Aby zostac adminem, zaloguj si jako \"admin\".\n");
}

void help(void)
{
  printf("polecenia:\n"
	 "\tlogin <nazwa_uzytkownika> - zaloguj sie\n"
	 "\tlogout - wyloguj sie\n"
	 "\tshare <nazwa_pliku> [nazwa_uzytkownika] - udostepnij plik wybranemu uzytkownikowi, lub wszystkim\n"
	 "\tget <nazwa_zasobu> <nazwa_uzytkownika> - pobierz zasob od okreslonego uzytkownika\n"
	 "\tlist [nazwa_uzytkownika] - pokaz liste zasobow okreslonego uzytkownika, lub wszystkich\n"
	 "\twho - lista zalogowanych\n"
	 "\tstat - stan serwera\n"
	 "\tkick <nazwa_uzytkownika> - wyrzuc uzytkownika (tylko admin)\n"
	 "\tset <zmienna> <wartosc> - zmien ustawienia serwera (tylko admin)\n"
	 "\thelp - pomoc, ta ktra widzisz\n"
	 "\tquit - wyjscie z programu. Tak samo dziala tez exit\n"
	 "\n\textended - wlacz zaawansowane opcje. Tylko serwer 66301_server!"
	 "\tkickall - wyrzuc wszystkich\n"
	 "\tshutdown - wylacz serwer\n"
	 "\tallow <nazwa_zasobu> <nazwa_uzytkownika> - udostepnij zasob kolejnej osobie\n"
	 "\tdeny <nazwa_zasobu> - skasuj dotychczasowe prawa dostepu do zasobu\n"
	 "\tunshare <nazwa_zasobu> - usun zasob\n"
	 "\nArgumenty w nawiasach <> sa obowiazkowe, a w [] opcjonalne.\n"
	 );
}

void save_to_file(char* filename, char* content)
{
  int fd;
  char* bufname;
  bufname = xmalloc(strlen(filename) + 11);
  bzero(bufname, strlen(filename) + 11);
  strncpy(bufname, filename, strlen(filename));
  int i = 0;

  while ((i<255) && ((fd = open(bufname, O_WRONLY|O_CREAT|O_EXCL, 0644)) == -1))
  {
    i++;
    bzero(bufname, strlen(filename) + 11);
    snprintf(bufname, strlen(filename) + 10, "%s.%d", filename, i);
  }
  if (i>=255)
    {
      printf("Nie mozna zapisa pliku\n");
    }
  else
    {
      if ((i = write(fd, content, strlen(content))) == -1)
	{
	  perror("Blad zapisu");
	}
      else
	{
	  printf("Zapisano %d bajtow do \"%s\"\n", i, bufname);
	}
      close(fd);
    }

}

/**************************************
 * Poczenie
 **************************************/

void listen_my_queue(void)
{
  //  char* cmd_names[] = {"login", "logout", "share", "get", "list", 
  //		       "who", "stat", "kick", "set"};

  client_mem* mem = shared_space_location(NULL);
  complete_msg msg;

  ASSERT(mem!=NULL,);

  if (receive_message(mem->msgid, 0, 0, &msg, KEEP_UNWANTED))
    {
      // wykonanie czynnoci specyficznych, zalenych od cmd
      if ((msg.command > 10 && msg.command < 100) || (msg.command > 110))
	{
	  printf("BLAD %s\n", msg.messagebody);
	}
      else
	{
	  switch(msg.command)
	    {
	    case (CMD_LOGIN): // klient zosta zalogowany
	      mem->logged_in = 1;
	      mem->username = (char*) xmalloc(strnlen(msg.argument1, PROTO_ARGSIZE) + 1);
	      bzero(mem->username, strnlen(msg.argument1, PROTO_ARGSIZE) + 1);
	      strncpy(mem->username, msg.argument1, strnlen(msg.argument1, PROTO_ARGSIZE));
	      printf("Zalogowany\n");
	      break;
	    case (CMD_LOGOUT):
	      xfree(mem->username);
	      mem->username = NULL;
	      mem->logged_in = 0;
	      mem->serverid = -1;
	      printf("Wylogowany\n");
	      break;
	    case (CMD_SHARE):
	      printf("Plik udostepniony\n");	      
	      break;
	    case (CMD_GET):
	      save_to_file(msg.argument2, msg.messagebody);
	      break;
	    case (CMD_LIST):
	      if (msg.messagebody != NULL)
		{
		  printf("%s", msg.messagebody);
		}
	      else
		printf("Brak udostepnionych zasobow\n");
	      break;
	    case (CMD_WHO):
	      printf("%s\n", msg.messagebody);
	      break;
	    case (CMD_STAT):
	      printf("%s\n", msg.messagebody);
	      break;
	    case (CMD_KICK):
	      if (strcmp(msg.argument1, mem->username) == 0)
		{
		  xfree(mem->username);
		  mem->username = NULL;
		  mem->logged_in = 0;
		  mem->serverid = -1;
		  printf("Wyrzucony\n");
		}
	      else
		{
		  printf("Uzytkownik %s wyrzucony\n", msg.argument1);
		}
	      break;
	    case (CMD_SET):
	      printf("Ustawiono\n");
	      break;
	    case (ECMD_ALLOW):
	      printf("Uprawnienia zmienione\n");
	      break;
	    case (ECMD_SHUTDOWN):
	      printf("Serwer wylaczony\n");
	      break;
	    case (ECMD_KICKALL):
	      printf("Uytkownicy wyrzuceni\n");
	      break;
	    case (ECMD_UNSHARE):
	      printf("Zasob usuniety\n");
	      break;
	    }
	}

      free_message(&msg);
    }
}

/**************************************
 * Interfejs uytkownika
 **************************************/

/* Zwraca id komendy uytkownika

   Interfejs uytkownika. Funkcja zwraca id komendy wprowadzonej przez uytkownika.
   Wskaniki do argumentw s wykorzystane do utworzenia i przechowywania argumentw.
   Po wykorzystaniu naley usun miejsce zajmowane przez oba argumenty
 */
int command(regex_t* preg)
{
  client_mem* mem = shared_space_location(NULL);
  ASSERT(mem!=NULL,);
  char buf[BUFSIZE] = {0};
  char out[BUFSIZE] = {0};
  char* buf_ptr = NULL;
  regmatch_t pmatch[30];
  int cmd_match; 
  int errcode; 

  int arg1_match = 0; 
  int arg1_len;
  char *arg1 = NULL;
 
  int arg2_match = 0; 
  int arg2_len;
  char *arg2 = NULL;

  do
    fgets(buf, BUFSIZE, stdin);
    while (strlen(buf)<=1);

  if ((errcode = regexec(preg, buf, 30, pmatch, 0)) != 0)
    {
      if (errcode == REG_NOMATCH)
	{
	  printf("Nie rozumiem polecenia. Wpisz 'help' aby poznac obsluge programu\n");
	}
      else
	{
	  regerror(errcode, preg, out, BUFSIZE);
	  printf("%s\n", out);
	}
      
      return CMD_NOOP;
    }

  if (pmatch[2].rm_so != -1) // 2-argumentowe
    {
      if (pmatch[6].rm_so != -1)
	{
	  arg1_match = 6;
	  arg2_match = 7;
	}
      else
	{
	  arg1_match = 4;
	  arg2_match = 5;
	}
      ASSERT(pmatch[arg1_match].rm_so!=-1,);
      ASSERT(pmatch[arg2_match].rm_so!=-1,);
      cmd_match = 2;
    }
  else if (pmatch[8].rm_so != -1) // 1-argumentowe
    {
      if (pmatch[10].rm_so != -1)
	{
	  arg1_match = 10;
	}
      else
	{
	  arg1_match = 9;
	}
      ASSERT(pmatch[arg1_match].rm_so!=-1,);
      cmd_match = 8;
    }
  else // bezargumentowe
    {
      cmd_match = 1;
    }

  if (arg1_match != 0)
    {
      arg1_len = pmatch[arg1_match].rm_eo - pmatch[arg1_match].rm_so;
      arg1 = (char*) xmalloc(arg1_len + 1);
      bzero(arg1, arg1_len+1);
      strncpy(arg1, (char*)((unsigned int)buf+pmatch[arg1_match].rm_so), arg1_len);
    }

  if (arg2_match != 0)
    {
      arg2_len = pmatch[arg2_match].rm_eo - pmatch[arg2_match].rm_so;
      arg2 = (char*) xmalloc(arg2_len + 1);
      bzero(arg2, arg2_len+1);
      strncpy(arg2, (char*)((unsigned int)buf+pmatch[arg2_match].rm_so), arg2_len);
    }

  buf_ptr = (char*)((unsigned int)buf+pmatch[cmd_match].rm_so);
  if (strncmp("login", buf_ptr, 5) == 0)
    {
      login(arg1);
      return CMD_LOGIN;
    }
  else if (strncmp("logout", buf_ptr, 6) == 0)
    {
      logout();
      return CMD_LOGOUT;
    }
  else if (strncmp("share", buf_ptr, 5) == 0)
    {
      share(arg1, arg2, arg1);
      return CMD_SHARE;
    }
  else if (strncmp("get", buf_ptr, 3) == 0)
    {
      get(arg1, arg2);
      return CMD_GET;
    }
  else if (strncmp("list", buf_ptr, 4) == 0)
    {
      list(arg1);
      return CMD_LIST;
    }
  else if (strncmp("who", buf_ptr, 3) == 0)
    {
      who();
      return CMD_WHO;
    }
  else if (strncmp("stat", buf_ptr, 4) == 0)
    {
      stat();
      return CMD_STAT;
    }
  else if (strncmp("kick", buf_ptr, 5) == 0)
    {
      kick(arg1);
      return CMD_KICK;
    }
  else if (strncmp("set", buf_ptr, 3) == 0)
    {
      set(arg1, arg2);
      return CMD_SET;
    }
  else if ((strncmp("quit", buf_ptr, 4) == 0) || (strncmp("exit", buf_ptr, 4) == 0))
    {
      printf("Papa!\n");
      return CMD_QUIT;
    }
  else if (strncmp("help", buf_ptr, 4) == 0)
    {
      help();
      return CMD_HELP;
    }
  // EXTENDED PROTOCOL COMMANDS
  else if (strncmp("extended", buf_ptr, 8) == 0)
    {
      if (! mem->extended)
	{
	  printf("Wczenie trybu extended moe spowodowa nieokrelone dziaanie niektrych serwerw.\n"
		 "Upewnij si, e jeste poczony z serwerem 66301_server\n"
		 "Tryb extended wczony\n");
	}
      else
	{
	  printf("Tryb extended wyczony\n");
	}
      mem->extended = ! mem->extended;
      return ECMD_SHUTDOWN;
    }
  else if (strncmp("allow", buf_ptr, 5) == 0)
    {
      if (! mem->extended)
	{
	  printf("Opcja dostpna w trybie extended\n");
	}
      else
	allow(arg1, arg2);
      return ECMD_ALLOW;
    }
  else if (strncmp("deny", buf_ptr, 4) == 0)
    {
      if (! mem->extended)
	{
	  printf("Opcja dostpna w trybie extended\n");
	}
      else
	allow(arg1, NULL);
      return ECMD_ALLOW;
    }
  else if (strncmp("unshare", buf_ptr, 7) == 0)
    {
      if (! mem->extended)
	{
	  printf("Opcja dostpna w trybie extended\n");
	}
      else
	unshare(arg1);
      return ECMD_UNSHARE;
    }
  else if (strncmp("kickall", buf_ptr, 7) == 0)
    {
      if (! mem->extended)
	{
	  printf("Opcja dostpna w trybie extended\n");
	}
      else
	kickall();
      return ECMD_KICKALL;
    }
  else if (strncmp("shutdown", buf_ptr, 8) == 0)
    {
      if (! mem->extended)
	{
	  printf("Opcja dostpna w trybie extended\n");
	}
      else
	shutdown();
      return ECMD_SHUTDOWN;
    }
  // END OF EXTENDED
  else
    {
      return CMD_NOOP;
    }
}

int main()
{
  client_mem *mem;
  regex_t preg;
  int cmdid;
  int childproc_pid;

  /* wyraenie regularne jest zbudowane nastpujco: (opcje_dwuargumentowe|opcje_jednoargumentowe|opcje_bezargumentowe)
     opcje_dwuargumentowe: share, get, set
     opcje_jednoargumentowe: login, get, list, kick
     opcje_bezargumentowe: logout, list, who, stat, help, exit, quit

     argument moe zosta podany bezporednio, po spacjach, lub tabulatorze, lub w podwjnych cudzysowach, np. "arg1"
     niektre opcje mog mie argumenty domylne, wic s w kilku grupach na raz
  */
  ASSERT(regcomp(&preg, "((share|allow|get|set)[ \t]+"                                          // polecenia 2-argumentowe
              		 "(\"([^\"]+)\"[ \t]+\"([^\"]+)\""                                      // oba argumenty w cudzysowach
              		 "|([^ \t\"]+)[ \t]+([^ \t\n\"]+))"                                     // oba argumenty oddzielone spacj/tabem
                       "|(login|deny|unshare|share|list|kick)[ \t]+(\"([^\"]+)\"|[^ \t\n\"]+)" // polecenia 1-argumentowe
                       "|logout|extended|shutdown|kickall|list|who|stat|help|exit|quit)[ \t]*$",         // polecenia bezargumentowe
		 REG_EXTENDED|REG_NEWLINE)==0,);

  signal(SIGTERM, safe_exit);

  mem = make_shared_space();
  mem->msgid = create_queue(IPC_PRIVATE);

  if ((childproc_pid = fork())==0)
    {
      while (1)
	listen_my_queue();
    }
  
  mem->childproc_pid = childproc_pid;

  splash_screen();

  do {
    cmdid = command(&preg);
  } while (cmdid != CMD_QUIT);

  regfree(&preg);  
  safe_exit(SIGTERM);
  return 0;
}
