/****
   LTPS - Linux Text Presentation System
                                       ****/

#include <iostream>
#include <unistd.h> // exit
#include <expat.h> // XML
#include <ncurses.h> // key codes

#include "slides.h"

#include "common.h"
#include "slide.h"

using namespace std;

// set this to 1 to see how tags are opened and closed
#define VERBOSE_XML_PARSING 0

#define META_NONE 0
#define META_AUTHOR 1
#define META_TITLE 2

typedef int metatag_t;

#define BUFSIZE 256

/* private struct for internal parsing state information */
struct xmlp_userdata {
  Slides* xml_DOM;
  metatag_t meta_tag; 
  Tag* tag_ptr;
  int position;
};

Slides::Slides(void)
{
  title = NULL;
  author = NULL;
}

Slides::Slides(char* filename)
{
  FILE *we_xml;
  char buf[BUFSIZE];
  int read_len;
  struct xmlp_userdata xmlp_ud;
  XML_Parser xmlp;

  title = NULL;
  author = NULL;

  if ((we_xml = fopen(filename, "r")) == NULL)
    {
      cout << "can't open file " << filename << endl;
      exit(1);
    }

  xmlp_ud.xml_DOM = this;
  xmlp_ud.tag_ptr = NULL;
  xmlp_ud.meta_tag = META_NONE;
  xmlp_ud.position = 0;

  xmlp = XML_ParserCreate(NULL);
  XML_SetUserData(xmlp, &xmlp_ud);
  XML_SetElementHandler(xmlp, &xml_opentag, &xml_closetag);
  XML_SetCharacterDataHandler(xmlp, &xml_chartag);

  while (! feof(we_xml))
    {
      read_len = fread(buf, sizeof(char), BUFSIZE, we_xml);
      if (! XML_Parse(xmlp, buf, read_len, feof(we_xml))) {
	cout << "xml file parse error: " << XML_ErrorString(XML_GetErrorCode(xmlp)) << endl;
	exit(1);
      }
    }

  fclose(we_xml);
  XML_ParserFree(xmlp);
}

Slides::~Slides(void)
{
}

void Slides::play(void)
{
  int slide_no, layer_no;
  int key, update=1;

  //  printf("%s prezentuje \"%s\"\n", getauthor(), gettitle());

  slide_no=0;
  layer_no=0;

  do {
    if (update)
      {
	key = slides[slide_no]->show(layer_no);
	update = 0;
      }
    else
      {
	key = getch();
      }

    switch (key)
      {
      case (' '): // NEXT SLIDE
	if (layer_no<slides[slide_no]->layers_count()-1 || slide_no<(signed)slides.size()-1)
	  {
	    update=1;
	    if (slide_no<(signed)slides.size()-1)
	      {
		layer_no = 0;
		slide_no++;
	      }
	    else
		layer_no++;
	  }
	break;
      case ('x'): // PREVIOUS SLIDE
	if (layer_no>0 || slide_no>0)
	  {
	    update=1;
	    if (slide_no>0)
	      {
		slide_no--;
		layer_no = slides[slide_no]->layers_count() -1;
	      }
	    else
		layer_no--;
	  }
	break;
      }
  } while (key != 'q');
}

void Slides::setauthor(char*newauthor, int len)
{
  if (this->author!=NULL)
    free(this->author);
  this->author = new char[len+1];
  bzero(this->author, len+1);
  strncpy(this->author, newauthor, len);
}

char* Slides::getauthor(void)
{
  return this->author;
}

void Slides::settitle(char*newtitle, int len)
{
  if (this->title!=NULL)
    free(this->title);
  this->title = new char[len+1];
  bzero(this->title, len+1);
  strncpy(this->title, newtitle, len);
}

char* Slides::gettitle(void)
{
  return this->title;
}


/* * * * * *
   PRIVATE
 * * * * * */

void xml_chartag(void *data, const XML_Char* s, int len)
{
  struct xmlp_userdata *xmldata = (struct xmlp_userdata*)data;
  char* buf1, *buf2;
  ASSERT(xmldata!=NULL, "xml_DOM has to point on an existing object");

  if (xmldata->meta_tag != META_NONE) // if we're in some meta tag
    {
      switch (xmldata->meta_tag)
	{
	case (META_AUTHOR): //  AUTHOR
	  buf1 = xmldata->xml_DOM->getauthor();
	  if (buf1==NULL)
	    xmldata->xml_DOM->setauthor((char*)s, len);
	  else
	    {
	      buf2 = new char[strlen(buf1) + len];
	      snprintf(buf2, strlen(buf1)+len, "%s%s", buf1, s);
	      xmldata->xml_DOM->setauthor(buf2, strlen(buf1)+len);
	    }
	  break;
	case (META_TITLE): // TITLE
	  buf1 = xmldata->xml_DOM->gettitle();
	  if (buf1==NULL)
	    xmldata->xml_DOM->settitle((char*)s, len);
	  else
	    {
	      buf2 = new char[strlen(buf1) + len];
	      snprintf(buf2, strlen(buf1)+len, "%s%s", buf1, s);
	      xmldata->xml_DOM->settitle(buf2, strlen(buf1)+len);
	    }
	  break;
	}
    }
  else if (xmldata->tag_ptr!=NULL) // if we're in some tag, pointed to by tag_ptr
    {
      xmldata->tag_ptr->appendcontent((char*)s, len);
    }
}

void xml_opentag(void *data, const char *el, const char **attr)
{
  struct xmlp_userdata *xmldata = (struct xmlp_userdata*)data;
  ASSERT(xmldata!=NULL, "xml_DOM has to point on an existing object");
  Tag* new_tag = NULL;

  // first of all check if we're in some particular top level tag (presentation/head/slides/slide)
  if (strcmp("presentation", el) == 0) // PRESENTATION
    {
      xmldata->position |= XML_IN_PRESENTATION;
    }
  else if (strcmp("head", el) == 0) // HEAD
    {
      xmldata->position |= XML_IN_HEAD;
    }
  else if (strcmp("slides", el) == 0) // SLIDES
    {
      xmldata->position |= XML_IN_SLIDES;
    }
  else if (strcmp("slide", el) == 0) // SLIDE
    {
      if (xmldata->position == (XML_IN_PRESENTATION|XML_IN_SLIDES))
	{
	  xmldata->position |= XML_IN_SLIDE;

	  Slide *newslide = new Slide(attr);
	  xmldata->xml_DOM->slides.push_back(newslide);
	  new_tag = xmldata->xml_DOM->slides[xmldata->xml_DOM->slides.size()-1];
	}
    } // or if we're in HEAD and have to add something there
  else if (xmldata->position & XML_IN_HEAD)
    {
      if (strcmp("author", el) == 0)
	{
	  xmldata->meta_tag = META_AUTHOR;
	}
      else if (strcmp("title", el) == 0)
	{
	  xmldata->meta_tag = META_TITLE;
	}
      
    } // or if we're in SLIDE and have to add a tag to the last added tag
  else if (xmldata->position & (XML_IN_SLIDE))
    {
      //      cout << "going to open " << el << " (in "<<xmldata->position<<") (parent: " << xmldata->tag_ptr << ")" << endl;
      new_tag = new Tag(xmldata->tag_ptr);
      if (new_tag->settype((char*)el) == -1)
	{
	  delete new_tag;
	  new_tag = NULL;
	  cout << "Unexpected <" << el << "> tag" << endl;
	}
      else
	{
	  xmldata->tag_ptr->addtag(new_tag);
	}
    }

  xmldata->tag_ptr = new_tag;

#if VERBOSE_XML_PARSING>1
  cout << "<" << el << " ptr=" << xmldata->tag_ptr << ">" << endl;
#endif
}

void xml_closetag(void *data, const char *el)
{
  struct xmlp_userdata *xmldata = (struct xmlp_userdata*)data;
  ASSERT(xmldata!=NULL, "xml_DOM has to point on an existing object");

  if (strcmp("presentation", el) == 0)
    {
      xmldata->position ^= XML_IN_PRESENTATION;
    }
  else if (strcmp("head", el) == 0)
    {
      xmldata->position ^= XML_IN_HEAD;
    }
  else if (strcmp("slides", el) == 0)
    {
      xmldata->position ^= XML_IN_SLIDES;
    }
  else if (strcmp("slide", el) == 0)
    {
      xmldata->position ^= XML_IN_SLIDE;      
      xmldata->tag_ptr = NULL;
    }
  else if (strcmp("author", el) == 0)
    {
      xmldata->meta_tag = META_NONE;
#if VERBOSE_XML_PARSING>1
      cout << "Autor prezentacji: " << xmldata->xml_DOM->getauthor() << endl;
#endif
    }
  else if (strcmp("title", el) == 0)
    {
      xmldata->meta_tag = META_NONE;
#if VERBOSE_XML_PARSING>1
      cout << "Tytu prezentacji: " << xmldata->xml_DOM->gettitle() << endl;
#endif
    }
  else
    {
      if (xmldata->tag_ptr != NULL)
	{
#if VERBOSE_XML_PARSING>1
	  cout << el << " (parent=" << xmldata->tag_ptr->getparent() << ") has " << xmldata->tag_ptr->tags.size() << " children" << endl;
#endif

	  if (xmldata->position & XML_IN_SLIDE)
	    {
	      xmldata->tag_ptr = xmldata->tag_ptr->getparent();
	    }
	}
      else
	{
	  cout << "SHIT " << el << " leaving with no pointer to it" << endl;
	}
    }

#if VERBOSE_XML_PARSING>1
  cout << "</" << el << ">" << endl;
#endif
}
